# Copyright 2017 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from __future__ import absolute_import

import fcntl
import gzip

import os
import sys
import time
import copy
import logging
import logging.handlers
from logging.handlers import SysLogHandler
from datetime import datetime
from json import JSONEncoder

import ipaddr
import socket
import yaml

from netq_lib.common.enums import BgColors, DefaultPaths, ConfKw


def color_wrap(color, text):
    # http://stackoverflow.com/questions/287871/
    # print-in-terminal-with-colors-using-python
    if color == BgColors.GREEN:
        return BgColors.GREEN + text + BgColors.ENDC
    elif color == BgColors.RED:
        return BgColors.RED + text + BgColors.ENDC
    elif color == BgColors.BLUE:
        return BgColors.BLUE + text + BgColors.ENDC
    elif color == BgColors.YELLOW:
        return BgColors.YELLOW + text + BgColors.ENDC
    elif color == BgColors.GRAY:
        return BgColors.GRAY + text + BgColors.ENDC
    elif color == BgColors.BOLD:
        return BgColors.BOLD + text + BgColors.ENDC
    else:
        raise RuntimeError('%s is not a supported color' % color)


def is_valid_ip(ip_address):
    """Returns True if this is a valid IPv4/v6 address"""
    try:
        if '/' in ip_address:
            _ = ipaddr.IPNetwork(ip_address)
        else:
            _ = ipaddr.IPAddress(ip_address)
        return True
    except Exception:
        return False


def is_valid_mac(mac_address):
    """Returns true if this is a valid MAC address"""
    return len(mac_address.split(':')) == 6


def pretty_date(input_time):
    """Get a datetime object or a int() Epoch timestamp and return a
    pretty string like 'an hour ago', 'Yesterday', '3 months ago',
    'just now', etc
    """
    if not input_time:
        return "never"

    now = datetime.utcnow()
    try:
        diff = now - datetime.utcfromtimestamp(input_time)
    except Exception:
        if isinstance(time, datetime):
            diff = now - input_time
        else:
            raise RuntimeError('Type of time (%s) is neither a timestamp or '
                               'datetime' % type(input_time))
    second_diff = diff.seconds
    day_diff = diff.days

    msg = None
    if day_diff < 0:
        msg = ''
    elif day_diff == 0:
        if second_diff < 2:
            msg = 'just now'
        elif second_diff < 60:
            msg = str(second_diff) + 's ago'
        elif second_diff < 120:
            msg = '1m ago'
        elif second_diff < 3600:
            msg = str(second_diff / 60) + 'm ago'
        elif second_diff < 7200:
            msg = '1h ago'
        elif second_diff < 86400:
            msg = str(second_diff / 3600) + 'h ago'
    elif day_diff == 1:
        msg = 'Yesterday'
    elif day_diff < 7:
        msg = str(day_diff) + 'd ago'
    elif day_diff < 31:
        msg = str(day_diff / 7) + 'W ago'
    elif day_diff < 365:
        msg = str(day_diff / 30) + 'M ago'
    else:
        msg = str(day_diff / 365) + 'Y ago'
    return msg


def strip_colors(text, colors=BgColors.COLORS):
    if not isinstance(text, basestring) or not colors:
        return text
    for color_idx, color in enumerate(colors):
        idx = text.find(color)
        if idx != -1:
            next_idx = text.find(BgColors.ENDC, idx)
            if next_idx != -1:
                return strip_colors(text[:idx] +
                                    text[idx + len(color):next_idx] +
                                    text[next_idx + len(BgColors.ENDC):],
                                    colors[color_idx + 1:])
    return text


def prtree(tree, shift, top, path, final_json):
    """print a tree when provided input in the form of:
    ['a', ['b', ['c', 'd']]]
    output looks like:
    a -- b -- c
           -- d
    STP Topo, MAC Path trace and others produce output in the form
    used as input by this routine
    """
    prefixstr = ''
    is_leaf_node = True
    host = ''
    port = ''
    entry = {}
    jsonify = False
    if final_json is not None:
        jsonify = True

    if 'node' in tree:
        fld = tree['node']
        if ':' in fld:
            words = fld.split(':')
            if words[1] == 'mac':
                host = words[0]
                port = ':'.join(words[1:])
            else:
                host = words[0]
                port = words[1]
        else:
            host = fld
            port = ''

    if jsonify:
        if 'node' in tree:
            for key in tree:
                if key == 'nexthop' or key == 'subtree':
                    continue    # We copy this separately
                if key == 'info' and tree[key] and 'Blocked' in tree[key]:
                    if 'nexthop' in tree:
                        entry['ingress'] = tree['nexthop']
                    else:
                        entry['ingress'] = port
                elif key == 'node':
                    pass
                else:
                    entry[key] = tree[key]
        if 'nexthop' in tree:
            if 'ingress' not in entry:
                entry['nexthop'] = tree['nexthop']
        elif port != '':
            entry['nexthop'] = port

        if 'node' in tree and tree['node']:
            entry['node'] = host

        path.extend([entry])
    else:
        if top:
            printstr = fld
            if 'info' in tree and tree['info']:
                if 'Error' in tree['info']:
                    printstr = color_wrap(BgColors.RED, printstr)
                elif 'Blocked' in tree['info']:
                    printstr = color_wrap(BgColors.GRAY,
                                          '%s:Blocked(%s)' % (host, port))
                elif 'ClagIsl' in tree['info']:
                    printstr = color_wrap(BgColors.YELLOW,
                                          '%s:ClagIsl(%s)' % (host, port))
                elif 'Root' in tree['info']:
                    printstr = color_wrap(BgColors.GREEN,
                                          'Root(%s)' % printstr)
                elif 'Edge' in tree['info']:
                    printstr = color_wrap(BgColors.BLUE,
                                          '%s:EdgeIntf(%s)' % (host, port))
                elif 'VxLAN L3' in tree['info']:
                    printstr = color_wrap(BgColors.GREEN, printstr)
                elif 'Blackhole' in tree['info']:
                    printstr = color_wrap(BgColors.GRAY, printstr)

            sys.stdout.write(prefixstr + printstr)
            shift += len(strip_colors(printstr))
            top = False
        else:
            printstr = ''
            if 'info' in tree and tree['info']:
                if 'Error' in tree['info']:
                    printstr = ' -- ' + color_wrap(BgColors.RED, fld)
                elif 'Blocked' in tree['info']:
                    printstr = color_wrap(BgColors.GRAY,
                                          ' -- %s:Blocked(%s)' % (host, port))
                elif 'ClagIsl' in tree['info']:
                    printstr = color_wrap(BgColors.YELLOW,
                                          ' -- %s:ClagIsl(%s)' % (host, port))
                elif 'Root' in tree['info']:
                    printstr = color_wrap(BgColors.GREEN,
                                          ' -- Root(%s)' % fld)
                elif 'Edge' in tree['info']:
                    printstr = color_wrap(BgColors.BLUE,
                                          ' -- %s:EdgeIntf(%s)' % (host, port))
                elif 'VxLAN L3' in tree['info']:
                    printstr = color_wrap(BgColors.GREEN, ' -- %s' % fld)
                elif 'Blackhole' in tree['info']:
                    printstr = color_wrap(BgColors.GRAY, ' -- %s' % fld)

            if 'vtep' in tree:
                printstr = color_wrap(BgColors.BLUE, ' -- %s' % fld)

            if printstr == '':
                printstr = ' -- ' + fld

            sys.stdout.write(prefixstr + printstr)
            shift += len(strip_colors(printstr))

    for fld in tree['subtree']:
        is_leaf_node = False
        if not jsonify:
            sys.stdout.write(prefixstr)
            new_path = []
        else:
            # We're going to branch here. Copy existing path to new
            new_path = copy.deepcopy(path)

        prtree(fld[0], shift, top, new_path, final_json)

        if not jsonify:
            prefixstr = '\n' + ' ' * shift

    if jsonify:
        # In JSON, we print the entire path at the final terminating node
        if is_leaf_node:
            final_json.append(path)
        return final_json
    else:
        return True


class ClassEncoder(JSONEncoder):

    def default(self, o):  # pylint: disable=method-hidden
        return o.__dict__


class Pidfile(object):
    """ Checks if another instance of the program is running.
    Explanation from http://stackoverflow.com/questions/220525/
    ensure-a-single-instance-of-an-application-in-linux
    The Right Thing is advisory locking using flock(LOCK_EX); in Python,
    this is found in the fcntl module.
    Unlike pidfiles, these locks are always automatically released when
    your process dies for any reason, have no race conditions exist
    relating to file deletion (as the file doesn't need to be deleted to
    release the lock), and there's no chance of a different process
    inheriting the PID and thus appearing to validate a stale lock.
    If you want unclean shutdown detection, you can write a marker
    (such as your PID, for traditionalists) into the file after
    grabbing the lock, and then truncate the file to 0-byte status
    before a clean shutdown (while the lock is being held); thus, if the
    lock is not held and the file is non-empty, an unclean shutdown is
    indicated.
    """
    # pylint: disable=missing-docstring
    def __init__(self, pidfile, procname, uuid=None):
        try:
            self.__pidfd = open(pidfile, 'ab+')
        except IOError as err:
            err.extra_info = 'Failed to open pidfile: %s' % pidfile
            raise
        self.__pidfile = pidfile
        self.__procname = procname
        self.__uuid = uuid

    def __str__(self):
        return self.__pidfile

    def is_running(self):
        pid = self.read()
        if pid is None:
            return False
        cmdline = '/proc/%s/cmdline' % pid
        try:
            with open(cmdline, 'r') as pidfd:
                exec_out = pidfd.readline()
            return self.__procname in exec_out and (not self.__uuid or
                                                    self.__uuid in exec_out)
        except IOError:
            return False

    def lock(self):
        try:
            fcntl.flock(self.__pidfd, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except IOError:
            raise RuntimeError('Another instance of this program is already '
                               'running.')

    def read(self):
        try:
            self.__pidfd.seek(0)
            return int(self.__pidfd.readline().strip())
        except Exception:  # pylint: disable=broad-except
            return

    def unlock(self):
        try:
            fcntl.flock(self.__pidfd, fcntl.LOCK_UN)
        except IOError:
            raise RuntimeError('Unable to unlock pid file.')

    def write(self, pid):
        self.__pidfd.truncate(0)
        self.__pidfd.write('%d\n' % pid)
        self.__pidfd.flush()


def update_backend(conf):
    """
    @sergey: Please update/remove this when you integrate your db stuff
    This is only for parity with current functionality
    """

    backend_dict = {ConfKw.SERVER: None,
                    'port': 6379}

    if not conf:
        conf = {ConfKw.BACKEND: backend_dict}
    elif not conf.get(ConfKw.BACKEND):
        conf[ConfKw.BACKEND] = backend_dict
    else:
        for key in backend_dict.keys():
            if key not in conf[ConfKw.BACKEND]:
                conf[ConfKw.BACKEND][key] = backend_dict[key]
    try:
        conf[ConfKw.BACKEND]['port'] = int(conf[ConfKw.BACKEND]['port'])
    except ValueError:
        raise RuntimeError(
            'Backend port %s must be an integer' % conf[ConfKw.BACKEND]['port'])

    server = conf[ConfKw.BACKEND][ConfKw.SERVER]
    if not server:
        raise RuntimeError(
            'Please specify backend server in config file')

    try:
        server = socket.getaddrinfo(server, 0)[0][4][0]
    except socket.gaierror:
        raise RuntimeError('Server %s is not known. Exiting...' % server)
    conf[ConfKw.BACKEND][ConfKw.SERVER] = server
    return conf


def update_config(conf_dict):
    try:
        dir_path = DefaultPaths.CONF_FILE.rsplit('/', 1)[0]
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)
        with open(DefaultPaths.CONF_FILE, 'w') as fp:
            fp.write('# See /usr/share/doc/netq/examples for full configuration file\n')
            yaml.dump(conf_dict, fp, default_flow_style=False, default_style='')
        print('Backend server config added to %s' %
              DefaultPaths.CONF_FILE)
    except Exception as ex:
        print('Failed to update backend server at %s' % DefaultPaths.CONF_FILE)
        return 1
    return 0


def get_yaml_config(proc_name=''):
    main_conf = dict()
    conf_files = []
    conf = dict()

    if os.path.exists(DefaultPaths.CONF_FILE):
        conf_files.append(DefaultPaths.CONF_FILE)

    for root, dirs, files in os.walk(DefaultPaths.CONF_DIR):
        for filename in sorted(files):
                conf_files.append(os.path.join(root, filename))

    for conf_file in conf_files:
        try:
            with open(conf_file, 'r') as fp:
                try:
                    conf = yaml.load(fp)
                except yaml.YAMLError as error:
                    raise RuntimeError(
                        'Unable to parse config file: %s, error: %s'
                        % (conf_file, str(error)))
        except IOError as error:
            raise RuntimeError('Unable to open config file: %s, error: %s' %
                               (conf_file, str(error)))
        if not conf:
            continue
        for key, value in conf.items():
            if not main_conf.get(key):
                main_conf[key] = value
            else:
                if isinstance(value, dict):
                    for key2, value2 in value.items():
                        main_conf[key][key2] = value2
                else:
                    main_conf[key] = value

    return main_conf


def get_logger(proc_name, conf=None):
    loglevel = 'info'
    loglevel_dict = {'debug': logging.DEBUG,
                     'info': logging.INFO,
                     'warning': logging.WARNING,
                     'error': logging.ERROR}
    if conf:
        proc_conf = conf.get(proc_name, None)
        if proc_conf:
            loglevel = conf[proc_name].get('log_level', 'info')
    logging_level = loglevel_dict.get(loglevel, logging.INFO)

    handler = SysLogHandler(address='/dev/log', facility=SysLogHandler.LOG_USER)
    formatter = logging.Formatter(proc_name +
                                  '[%(process)d]: %(levelname)s: %(message)s')
    handler.setLevel(logging_level)
    handler.setFormatter(formatter)
    logger = logging.getLogger(proc_name)
    logger.setLevel(logging_level)
    logger.addHandler(handler)
    return logger


def list_to_ranges(in_list):
    '''Convert a list into a compact hyphenated string'''
    # This routine was copied from stackoverflow:
    # http://stackoverflow.com/questions/9847601/convert-list-of-numbers-to-string-ranges
    ret = []
    for val in sorted(in_list):
        if not ret or ret[-1][-1] + 1 != val:
            ret.append([val])
        else:
            ret[-1].append(val)
    return ",".join([str(x[0])
                     if len(x) == 1 else str(x[0]) + "-" + str(x[-1])
                     for x in ret])
