# Copyright 2016, 2017 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

import itertools
import sys
import time

import redis
from redis import WatchError
from redis.client import StrictPipeline, StrictRedis, PubSub, PubSubWorkerThread
from redis._compat import iteritems

from redis.exceptions import (
    ConnectionError,
    NoScriptError,
    ResponseError,
    TimeoutError,
    PubSubError
)

_RETRY_LIMIT = 6
_SOCKET_TIMEOUT = 5             # 5 seconds is quite long

class RedisException(Exception):

    def __init__(self, *args, **kwargs):
        super(RedisException, self).__init__(*args, **kwargs)

BackendExceptions = (ConnectionError, NoScriptError, ResponseError,
                     TimeoutError, PubSubError, RedisException)


class MyStrictRedis(StrictRedis):

    def execute_command(self, *args, **options):
        """Execute a command and return a parsed response
        """
        pool = self.connection_pool
        command_name = args[0]
        timeout = 1
        for retry in range(_RETRY_LIMIT + 1):
            connection = pool.get_connection(command_name, **options)
            try:
                connection.send_command(*args)
                output = self.parse_response(connection, command_name,
                                             **options)
                return output
            except (ConnectionError, TimeoutError, ResponseError) as ex:
                connection.disconnect()
                if isinstance(ex, NoScriptError):
                    raise
                if retry == _RETRY_LIMIT:
                    exc_type, value, traceback = sys.exc_info()
                    raise RedisException, ('execute_command failed',
                                           exc_type, value), traceback
            finally:
                try:
                    pool.release(connection)
                except: # pylint: disable=bare-except
                    pass
            timeout *= 2
            time.sleep(timeout)


class MyStrictPipeline(StrictPipeline):

    def execute(self, raise_on_error=True):
        """Execute all the commands in the current pipeline
        ** EVENTLET FRIENDLY VERSION ***
        """
        stack = self.command_stack
        if not stack:
            return []
        if self.transaction or self.explicit_transaction:
            execute = self._execute_transaction
        else:
            execute = self._execute_pipeline
        timeout = 1
        for retry in range(_RETRY_LIMIT + 1):
            conn = (
                self.connection_pool.get_connection('MULTI',
                                                    self.shard_hint)
            )
            try:
                if self.scripts:
                    self.load_scripts()
                output = execute(conn, stack, raise_on_error)
                self.scripts = set() # pylint: disable=attribute-defined-outside-init
                return output
            except (ConnectionError, TimeoutError, ResponseError):
                conn.disconnect()
                if retry == _RETRY_LIMIT:
                    exc_type, value, traceback = sys.exc_info()
                    raise RedisException, ('pipeline execute failed',
                                           exc_type, value), traceback
                # if we were watching a variable, the watch is no longer valid
                # since this connection has died. raise a WatchError, which
                # indicates the user should retry his transaction. If this is
                # more than a temporary failure, the WATCH that the user next
                # issues will fail, propagating the real ConnectionError
                if self.watching:
                    raise WatchError('A ConnectionError occurred on while '
                                     'watching one or more keys')
            finally:
                self.connection = conn # pylint: disable=attribute-defined-outside-init
                self.my_reset()
            # In a monkey patched environment, eventlet monkey patches
            # time.sleep by default.
            timeout *= 2
            time.sleep(timeout)

    def immediate_execute_command(self, *args, **options):
        """Execute a command immediately, but don't auto-retry on a
        ConnectionError if we're already WATCHing a variable. Used when
        issuing WATCH or subsequent commands retrieving their values but before
        MULTI is called.
        ** EVENTLET FRIENDLY VERSION ***
        """
        command_name = args[0]
        timeout = 1
        for retry in range(_RETRY_LIMIT + 1):
            conn = (
                self.connection_pool.get_connection(command_name,
                                                    self.shard_hint)
            )
            try:
                conn.send_command(*args)
                output = self.parse_response(conn, command_name, **options)
                self.command_stack = [] # pylint: disable=attribute-defined-outside-init
                self.scripts = set() # pylint: disable=attribute-defined-outside-init
                return output
            except (ConnectionError, TimeoutError, ResponseError) as ex:
                conn.disconnect()
                if isinstance(ex, NoScriptError):
                    raise
                if retry == _RETRY_LIMIT:
                    exc_type, value, traceback = sys.exc_info()
                    raise RedisException, ('pipeline immediate_execute failed',
                                           exc_type, value), traceback
                if self.watching:
                    raise WatchError('A ConnectionError occurred on while '
                                     'watching one or more keys')
            finally:
                self.connection = conn # pylint: disable=attribute-defined-outside-init
                self.my_reset()
            # In a monkey patched environment, eventlet monkey patches
            # time.sleep by default.
            timeout *= 2
            time.sleep(timeout)

    def load_scripts(self):
        # make sure all scripts that are about to be run on this pipeline exist
        scripts = list(self.scripts)
        immediate = self.immediate_execute_command
        shas = {s.sha for s in scripts}
        # we can't use the normal script_* methods because they would just
        # get buffered in the pipeline.
        exists = immediate('SCRIPT', 'EXISTS', *shas, **{'parse': 'EXISTS'})
        if not all(exists):
            for scr, exist in itertools.izip(scripts, exists):
                if not exist:
                    scr.sha = immediate('SCRIPT', 'LOAD', scr.script,
                                        **{'parse': 'LOAD'})

    def my_reset(self):
        # make sure to reset the connection state in the event that we were
        # watching something
        if self.watching and self.connection: # pylint: disable=access-member-before-definition
            try:
                # call this manually since our unwatch or
                # immediate_execute_command methods can call reset()
                self.connection.send_command('UNWATCH')
                self.connection.read_response()
            except ConnectionError:
                # disconnect will also remove any previous WATCHes
                self.connection.disconnect()
        # clean up the other instance attributes
        self.watching = False # pylint: disable=attribute-defined-outside-init
        self.explicit_transaction = False # pylint: disable=attribute-defined-outside-init
        # we can safely return the connection to the pool here since we're
        # sure we're no longer WATCHing anything
        if self.connection:
            try:
                self.connection_pool.release(self.connection)
            except: # pylint: disable=bare-except
                pass
            self.connection = None # pylint: disable=attribute-defined-outside-init


# This code is based on the original PubSub and PubSubWorkerThread classes at
# https://github.com/andymccurdy/redis-py/blob/2.10.5/redis/client.py.
# PubSubWorkerThread.run() doesnt have any exception handling, which leads
# to worker threads throwing an exception in the logs. NetQPubSubWorkerThread
# handles it.

class NetQPubSub(PubSub):

    def run_in_thread(self, sleep_time=0):
        for channel, handler in iteritems(self.channels):
            if handler is None:
                raise PubSubError("Channel: '%s' has no handler registered")
        for pattern, handler in iteritems(self.patterns):
            if handler is None:
                raise PubSubError("Pattern: '%s' has no handler registered")

        thread = NetQPubSubWorkerThread(self, sleep_time)
        thread.start()
        return thread


class NetQPubSubWorkerThread(PubSubWorkerThread):

    def run(self):
        if self._running:
            return
        self._running = True
        pubsub = self.pubsub
        sleep_time = self.sleep_time
        try:
            while pubsub.subscribed:
                pubsub.get_message(ignore_subscribe_messages=True,
                                   timeout=sleep_time)
        except (ConnectionError, AttributeError):
            pass
        pubsub.close()
        self._running = False


class Connection(object):

    PIPE = None
    REDIS = None
    SERVER = None
    PUBSUB = None
    CHANNELS = None
    LISTENER = None
    AGENT = False
    SUBSCRIBE = False

    def __init__(self):
        raise NotImplementedError

    @classmethod
    def get_pipeline(cls):
        return cls.PIPE

    @classmethod
    def new_pipeline(cls):
        cls.PIPE = MyStrictPipeline(cls.REDIS.connection_pool,
                                    cls.REDIS.response_callbacks,
                                    transaction=False,
                                    shard_hint=None)

    @classmethod
    def new_redis(cls):
        cls.REDIS = MyStrictRedis(cls.SERVER,
                                  socket_connect_timeout=_SOCKET_TIMEOUT,
                                  socket_timeout=_SOCKET_TIMEOUT)

    @classmethod
    def new_strict_redis(cls):
        cls.REDIS = StrictRedis(cls.SERVER,
                                socket_connect_timeout=_SOCKET_TIMEOUT,
                                socket_timeout=_SOCKET_TIMEOUT)

    @classmethod
    def subscribe(cls):
        if cls.CHANNELS:
            cls.PUBSUB = NetQPubSub(cls.REDIS.connection_pool)
            cls.PUBSUB.psubscribe(**cls.CHANNELS)
            cls.LISTENER = cls.PUBSUB.run_in_thread(sleep_time=0.01)

    @classmethod
    def unsubscribe(cls):
        if cls.LISTENER:
            cls.LISTENER.stop()
        if cls.PUBSUB:
            cls.PUBSUB.close()


def curr_pipeline():
    return Connection.PIPE


def new_pipeline():
    Connection.new_pipeline()


def are_conn_and_listener_alive():
    try:
        Connection.REDIS.ping()
    except BackendExceptions:
        return False
    return Connection.LISTENER.is_alive()


def unsubscribe_listener():
    try:
        Connection.unsubscribe()
    except BackendExceptions:
        pass


def ping_server(server=None):
    if server is None:
        server = Connection.SERVER
    redis.StrictRedis(server, socket_timeout=30,
                      socket_connect_timeout=2).ping()


def server_info(server=None):
    # return redis server info as a dict for caller to process
    if server is None:
        server = Connection.SERVER
    return redis.StrictRedis(server, socket_timeout=30,
                             socket_connect_timeout=2).info()


def reset_connection_settings():
    Connection.unsubscribe()
    Connection.PIPE.reset()
    if not Connection.AGENT:
        Connection.new_strict_redis()
    else:
        Connection.new_redis()
    Connection.new_pipeline()
    if Connection.SUBSCRIBE:
        Connection.subscribe()


def close_connection():
    if Connection:
        Connection.unsubscribe()
        Connection.PIPE.reset()


def set_connection_settings(backend, **kwargs):
    """
    Establish connection with Redis/Sentinels

    This method sets up connection to a single Redis server or
    a group of Sentinels.

    When ``backend`` has a single item(dict) or a list with a single
    item THEN it is a sigle master mode.
    When ``backend`` is a list(of dict) then it is a master-sentinel
    mode.

    Args:
        backend: Backend from configuration file.
    Kwargs:
        agent (bool):
        subscribe (bool):
        channels_dict (dict):

    Returns:
        (server, port) tuple that conneciton was established with.
    """
    server, port = None, None
    if not backend:
        raise ValueError('Backend is empty')

    if isinstance(backend, list):
        if len(backend) == 1:
            backend = backend[0]
        else:
            raise NotImplementedError('Sentinel support is not enabled.')

    if isinstance(backend, dict):
        server = backend.get('server', None)
        port = backend.get('port', None)

    if not server:
        raise ValueError('Server is empty')

    agent = kwargs.pop('agent', False)
    subscribe = kwargs.pop('subscribe', False)

    Connection.SERVER = server.split('/')[0]
    Connection.AGENT = agent
    Connection.SUBSCRIBE = subscribe
    Connection.CHANNELS = kwargs.pop('channels_dict', None)
    if not agent:
        Connection.new_strict_redis()
    else:
        Connection.new_redis()
    Connection.new_pipeline()
    if subscribe:
        Connection.subscribe()

    return server, port
