# Copyright 2016, 2017 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

"""ROM models
"""
import inspect
import time

from netq_lib.orm import objects, HOSTNAME
from netq_lib.orm.memdb.models import MemModel
from netq_lib.orm.redisdb import Connection
from netq_lib.orm.redisdb.query import RedisQuery

# If you add any classes to this file that is not an object pushed to the
# backend DB, such as  _RedisModel and Connection, please add those classes
# to the ignored list in agent.py's reinitialize_db() function.

_KEY_TTL = 90                   # TTL liveness is 90s

class _RedisModel(objects.NetQModel):
    __memmodel = None
    query_cls = RedisQuery
    active = objects.Boolean(json_key='active')
    hostname = objects.String(json_key='hostname')
    timestamp = objects.Float(json_key='timestamp')
    deleted = objects.Boolean(json_key='deleted')

    def __init__(self):
        self.active = True
        self.deleted = False
        self.hostname = HOSTNAME
        self.timestamp = time.time()
        super(_RedisModel, self).__init__()

    @staticmethod
    def _get_connection_object(commit):
        if commit:
            return Connection.REDIS
        else:
            return Connection.PIPE

    def copy(self, src):
        for _, field in inspect.getmembers(self.__class__):
            if (isinstance(field, objects.Field) and
                    field.name != self.__class__.timestamp.name):
                src_attribute = getattr(src, field.name, None)
                if src_attribute is not None:
                    setattr(self, field.name, src_attribute)

    def delete(self, purge=False):
        if purge:
            score = self.__class__.obj_to_score(self.__dict__)
            Connection.PIPE.zremrangebyscore(
                self.__class__.obj_to_key(self.__dict__),
                min=score,
                max=score
            )
            Connection.PIPE.hdel(self.__class__,
                                 self.__class__.obj_to_key(self.__dict__))

        else:
            obj = self.__class__()
            obj.copy(self)
            obj.active = False
            obj.deleted = True
            obj.save()

    @classmethod
    def exclude_fmt(cls):
        """Return the list of fields to ignore when doing a comparison to
        find the differences between two instances of an object.
        See BgpSession for an example usage"""
        return cls.timestamp, cls.active

    @classmethod
    def memmodel(cls):
        if cls.__memmodel is None:
            cls.__memmodel = type(cls.__name__, (MemModel, cls), {})
        return cls.__memmodel

    def save_entry(self, connection, publish=False):

        cls = self.__class__
        mcls = cls.memmodel()
        key = cls.obj_to_key(self.__dict__)
        val = cls.obj_to_val(self.__dict__)
        score = cls.obj_to_score(self.__dict__)

        add = True
        delete = False
        old_obj = None

        if mcls is not None:
            old_objs = (
                list(mcls.query.range(start=0, end=10000, first=False,
                                      keys=[key]))
            )
            if old_objs:
                old_obj = old_objs[-1]

                if self.equals(old_obj):
                    if old_obj.active == self.active:
                        add = False
                    elif self.active and len(old_objs) >= 2:
                        delete = True

                for obj in old_objs[:-2]:
                    obj.delete_object()

        if delete:
            del_score = cls.obj_to_score(old_obj.__dict__)
            mcls.delete(key, del_score)
            connection.zremrangebyscore(key, del_score, del_score)
        elif add:
            mcls.save(key, score, val)
            connection.zadd(key, score, val)
            connection.hset(cls.SEPERATOR.join((
                cls._namespace, cls.__name__)),
                key, self.active)
            if publish:
                connection.publish(key, val)

    @classmethod
    def key_fmt(cls, key_type=None):
        raise NotImplementedError

    @classmethod
    def score_fmt(cls, key_type=None):
        raise NotImplementedError

    @classmethod
    def val_fmt(cls):
        raise NotImplementedError

    def diff(self, other):
        '''
        Returns dict { k:(self,other) } differences between self and other instance same class
        Only finds deltas between common keys
        '''

        if other is None:
            raise ValueError

        if not type(self) is type(other):
            raise NotImplementedError

        #WARNING: Using PREFIX from objects.
        exclude_fields = {'i_'+field.name for field in self.exclude_fmt()}
        return { k:(v, other.__dict__.get(k)) for k,v in self.__dict__.items() if k not in exclude_fields and other.__dict__.get(k) != v }


class Address(_RedisModel):
    ifname = objects.String(json_key='ifName')
    prefix = objects.String(json_key='prefix')
    mask = objects.Integer(json_key='mask')
    is_ipv6 = objects.Boolean(json_key='isIPv6')
    vrf = objects.String(json_key='vrf')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.ifname, cls.prefix,
                cls.mask, cls.is_ipv6, cls.vrf)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return cls.timestamp, cls.active, cls.deleted

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Address, self).save_entry(connection)


class BgpSession(_RedisModel):
    """Store BGP Session specific info
    """
    peer_name = objects.String(json_key='peerName')
    state = objects.String(json_key='peerState')
    peer_router_id = objects.String(json_key='bgpRouterId')
    peer_asn = objects.Long(json_key='peerAsn')
    peer_hostname = objects.String(json_key='peerHostname')
    asn = objects.Long(json_key='bgpAsn')
    reason = objects.String(json_key='reason')
    ipv4_pfx_rcvd = objects.Integer(json_key='ipv4PrefixRx')
    ipv6_pfx_rcvd = objects.Integer(json_key='ipv6PrefixRx')
    last_reset_time = objects.Float(json_key='lastResetTime')
    up_time = objects.Float(json_key='upTime')
    conn_estd = objects.Integer(json_key='connEstd')
    conn_dropped = objects.Integer(json_key='connDropped')
    upd8_rx = objects.Integer(json_key='upd8Rx')
    upd8_tx = objects.Integer(json_key='upd8Tx')
    vrfid = objects.Integer(json_key='vrfId')
    vrf = objects.String(json_key='vrf')

    @classmethod
    def exclude_fmt(cls):
        return (cls.timestamp, cls.active, cls.last_reset_time, cls.up_time,
                cls.upd8_tx, cls.upd8_rx)

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.peer_name, cls.asn, cls.vrf

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.state, cls.peer_router_id, cls.peer_asn, cls.peer_hostname,
                cls.reason, cls.ipv4_pfx_rcvd, cls.ipv6_pfx_rcvd,
                cls.timestamp, cls.last_reset_time, cls.conn_estd,
                cls.conn_dropped, cls.upd8_rx, cls.vrfid, cls.upd8_tx,
                cls.up_time, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(BgpSession, self).save_entry(connection, publish=True)


class ClagSession(_RedisModel):
    role = objects.String(json_key='role')
    peer_role = objects.String(json_key='peerRole')
    peer_state = objects.Boolean(json_key='peerState')
    peer_if = objects.String(json_key='clagPeerIf')
    clag_sysmac = objects.String(json_key='sysMac')
    backup_ip_active = objects.Boolean(json_key='backupIpActive')
    backup_ip = objects.String(json_key='backupIp')
    single_bonds = objects.Json(json_key='singleBonds')
    dual_bonds = objects.Json(json_key='dualAttachedBonds')
    conflicted_bonds = objects.Json(json_key='conflictedBonds')
    proto_down_bonds = objects.Json(json_key='protoDownBonds')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.clag_sysmac

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.peer_role, cls.role, cls.peer_state, cls.peer_if,
                cls.backup_ip, cls.backup_ip_active, cls.single_bonds,
                cls.dual_bonds, cls.timestamp, cls.conflicted_bonds,
                cls.proto_down_bonds, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(ClagSession, self).save_entry(connection, publish=True)

class LnvSession(_RedisModel):
    role = objects.String(json_key='role')
    snd_ip = objects.String(json_key='sndIp')
    repl_mode = objects.String(json_key='replicationMode')
    rd_peers = objects.String(json_key='vxrdPeers')
    snd_peers = objects.Json(json_key='vxsndPeers')
    vnis = objects.Json(json_key='vnis')
    state = objects.String(json_key='state')
    version = objects.String(json_key='version')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.role

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.role, cls.snd_ip, cls.rd_peers, cls.snd_peers,
                cls.vnis, cls.state, cls.repl_mode, cls.version,
                cls.active, cls.deleted, cls.timestamp)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(LnvSession, self).save_entry(connection, publish=True)


class Command(_RedisModel):
    key = objects.String(json_key='key')
    output = objects.String(json_key='output')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.key

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return cls.output,

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        connection.zadd(Command.obj_to_key(self.__dict__),
                        Command.obj_to_score(self.__dict__),
                        Command.obj_to_val(self.__dict__))
        connection.hset(self.__class__.SEPERATOR.join((
            self.__class__._namespace,
            self.__class__.__name__)),
            Command.obj_to_key(self.__dict__),
            True)


class Link(_RedisModel):
    ifname = objects.String(json_key='ifName')
    ifindex = objects.Integer(json_key='ifIndex')
    mac_address = objects.MacAddress(json_key='macAddress')
    admin_state = objects.String(json_key='adminState')
    oper_state = objects.String(json_key='operState')
    managed = objects.Boolean(json_key='managed')
    kind = objects.String(json_key='kind')
    mtu = objects.Integer(json_key='mtu')
    master = objects.String(json_key='master')
    is_vlan_filtering = objects.Boolean(json_key='isVlanFiltering')
    vlans = objects.String(json_key='vlans')
    access_vlan = objects.Integer(json_key='accessVlan')
    dstport = objects.Integer(json_key='dstPort')
    localip = objects.String(json_key='localIp')
    vni = objects.Integer(json_key='vni')
    vrf = objects.String(json_key='vrf')
    rt_table_id = objects.Integer(json_key='routeTableId')
    parent_if = objects.String(json_key='parentIf')
    stp_state = objects.Integer(json_key='stpState')

    @classmethod
    def key_fmt(cls, key_type=None):
        """We need to query sometimes by VNI and so we stick VNI in the key.
        Its 0 for anything but Vxlan.
        """
        return (cls.__name__, cls.hostname, cls.ifname, cls.kind, cls.vni,
                cls.master)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.admin_state, cls.oper_state, cls.managed, cls.mtu,
                cls.ifindex, cls.is_vlan_filtering, cls.timestamp,
                cls.vlans, cls.access_vlan, cls.localip,
                cls.vrf, cls.rt_table_id, cls.parent_if, cls.stp_state,
                cls.mac_address, cls.dstport, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Link, self).save_entry(connection, publish=True)


class Liveness(_RedisModel):

    def __str__(self):
        return self.hostname

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return cls.hostname,

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        key = Liveness.obj_to_key(self.__dict__)
        val = Liveness.obj_to_val(self.__dict__)
        score = Liveness.obj_to_score(self.__dict__)
        connection.zadd(key, score, val)
        connection.expire(key, _KEY_TTL)
        connection.hset(self.__class__.SEPERATOR.join((
            self.__class__._namespace,
            self.__class__.__name__)),
            key, True)


class Lldp(_RedisModel):
    ifname = objects.String(json_key='ifname')
    peer_hostname = objects.String(json_key='peerHostname')
    peer_ifname = objects.String(json_key='peerIfname')
    lldp_peer_bridge = objects.Boolean(json_key='lldpPeerBridge')
    lldp_peer_router = objects.Boolean(json_key='lldpPeerRouter')
    lldp_peer_station = objects.Boolean(json_key='lldpPeerStation')
    lldp_peer_os = objects.String(json_key='lldpPeerOS')
    lldp_peer_osv = objects.String(json_key='lldpPeerOsVersion')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.ifname, cls.peer_hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.peer_ifname, cls.lldp_peer_bridge, cls.lldp_peer_router,
                cls.lldp_peer_station, cls.lldp_peer_os, cls.lldp_peer_osv,
                cls.timestamp, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Lldp, self).save_entry(connection, publish=True)


class MacFdb(_RedisModel):
    mac_address = objects.MacAddress(json_key='macAddress')
    origin = objects.Boolean(json_key='origin')
    nexthop = objects.String(json_key='nextHop')
    vlan = objects.Integer(json_key='vlan')
    dst = objects.String(json_key='dst')
    port = objects.String(json_key='port')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.mac_address, cls.vlan

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.origin, cls.nexthop, cls.dst, cls.port, cls.timestamp,
                cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(MacFdb, self).save_entry(connection)


class VxlanRemoteDest(_RedisModel):
    vni = objects.Integer(json_key='vni')
    rdst = objects.String(json_key='remoteDest')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.vni, cls.rdst

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.vni, cls.rdst, cls.active, cls.deleted, cls.timestamp)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(VxlanRemoteDest, self).save_entry(connection, publish=True)


class MstpInfo(_RedisModel):
    """Store MSTP non-port-specific info for troubleshooting
    """
    bridge_name = objects.String(json_key='bridgeName')
    state = objects.Boolean(json_key='state')
    root_port_name = objects.String(json_key='rootPortName')
    root_bridge = objects.String(json_key='rootBridge')
    topo_chg_ports = objects.Json(json_key='topoChangePort')
    time_since_tcn = objects.Long(json_key='timeSinceLastTcn')
    topo_chg_cntr = objects.Long(json_key='toptChangeCntr')
    bridge_id = objects.String(json_key='bridgeId')
    edge_ports = objects.Json(json_key='edgePorts')
    network_ports = objects.Json(json_key='networkPorts')
    disputed_ports = objects.Json(json_key='disputedPorts')
    bpduguard_ports = objects.Json(json_key='bpduguardPorts')
    bpduguard_err_ports = objects.Json(json_key='bpduguardErrPorts')
    ba_inconsistent_ports = objects.Json(json_key='baInconsistentPorts')
    bpdufilter_ports = objects.Json(json_key='bpdufilterPorts')
    is_vlan_filtering = objects.Boolean(json_key='isVlanFiltering')
    ports = objects.Json(json_key='ports')

    @classmethod
    def exclude_fmt(cls):
        return cls.timestamp, cls.active, cls.time_since_tcn

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.bridge_name

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.root_port_name, cls.topo_chg_ports, cls.time_since_tcn,
                cls.topo_chg_cntr, cls.ports, cls.edge_ports, cls.state,
                cls.network_ports, cls.disputed_ports, cls.bpduguard_ports,
                cls.root_bridge, cls.bridge_id, cls.bpduguard_err_ports,
                cls.ba_inconsistent_ports, cls.bpdufilter_ports,
                cls.is_vlan_filtering, cls.active, cls.deleted,
                cls.timestamp)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(MstpInfo, self).save_entry(connection, publish=True)


class OspfNbr(_RedisModel):
    peer_id = objects.String(json_key='peerId')
    peer_addr = objects.String(json_key='peerAddress')
    state = objects.String(json_key='state')
    ifname = objects.String(json_key='ifName')
    is_ipv6 = objects.String(json_key='isIPv6')
    area = objects.String(json_key='area')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.ifname, cls.peer_id, cls.area

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.state, cls.timestamp, cls.ifname, cls.is_ipv6,
                cls.peer_addr, cls.area, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(OspfNbr, self).save_entry(connection, publish=True)


class OspfIf(_RedisModel):
    ifname = objects.String(json_key='ifName')
    is_unnumbered = objects.Boolean(json_key='isUnnumbered')
    if_up = objects.Boolean(json_key='ifUp')
    area = objects.String(json_key='area')
    network_type = objects.String(json_key='networkType')
    nbr_count = objects.Integer(json_key='nbrCount')
    nbr_adj_count = objects.Integer(json_key='nbrAdjCount')
    cost = objects.Integer(json_key='cost')
    is_passive = objects.Boolean(json_key='isPassive')
    mtu = objects.Integer(json_key='mtu')
    dead_time = objects.Integer(json_key='deadTime')
    rexmit_time = objects.Integer(json_key='rexmitTime')
    hello_time = objects.Integer(json_key='helloTime')
    router_id = objects.String(json_key='routerId')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.ifname, cls.area

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.network_type, cls.timestamp, cls.nbr_count, cls.if_up,
                cls.nbr_adj_count, cls.is_unnumbered, cls.is_passive, cls.cost,
                cls.mtu, cls.dead_time, cls.rexmit_time, cls.hello_time,
                cls.router_id, cls.area, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(OspfIf, self).save_entry(connection, publish=True)


class Neighbor(_RedisModel):
    ifname = objects.String(json_key='ifName')
    ifindex = objects.Integer(json_key='ifIndex')
    ip_address = objects.String(json_key='ipAddress')
    mac_address = objects.MacAddress(json_key='macAddress')
    is_ipv6 = objects.Boolean(json_key='isIPv6')
    vrf = objects.String(json_key='vrf')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.ifname,
                cls.ip_address, cls.mac_address, cls.is_ipv6, cls.vrf)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return cls.ifindex, cls.timestamp, cls.active, cls.deleted

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Neighbor, self).save_entry(connection)


class Node(_RedisModel):
    lastboot = objects.Float(json_key='lastBoot')
    sys_uptime = objects.Float(json_key='sysUptime')

    def __str__(self):
        return self.hostname

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return cls.lastboot, cls.sys_uptime

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        key = Node.obj_to_key(self.__dict__)
        score = Node.obj_to_score(self.__dict__)
        val = Node.obj_to_val(self.__dict__)
        connection.zadd(key, score, val)
        connection.hset(self.__class__.SEPERATOR.join((
            self.__class__._namespace,
            self.__class__.__name__)),
            key, True)


class Route(_RedisModel):
    prefix = objects.String(json_key='prefix')
    protocol = objects.String(json_key='protocol')
    route_type = objects.Integer(json_key='routeType')
    origin = objects.Boolean(json_key='origin')
    nexthops = objects.Json(json_key='nextHops')
    src = objects.String(json_key='src')
    is_ipv6 = objects.Boolean(json_key='isIPv6')
    rt_table_id = objects.Integer(json_key='routeTableId')
    vrf = objects.String(json_key='vrf')

    @classmethod
    def key_fmt(cls, key_type=None):
        # The exact sequence of key fields here is assumed in the lpm LUA
        # script. Do not change one without changing the other. We need both
        # VRF and table in key to handle redistribute neighbor table. Also names
        # are constant across nodes, but table ids may vary from box to box. So,
        # save the routes with the VRF name as well as table id.
        return (cls.__name__, cls.hostname, cls.prefix, cls.route_type,
                cls.rt_table_id, cls.is_ipv6, cls.origin, cls.protocol, cls.vrf)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.nexthops, cls.src, cls.timestamp, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Route, self).save_entry(connection)


class Services(_RedisModel):
    name = objects.String(json_key='name')
    is_enabled = objects.Boolean(json_key='isEnabled')
    is_active = objects.Boolean(json_key='isActive')
    is_monitored = objects.Boolean(json_key='isMonitored')
    status = objects.Integer(json_key='status')
    vrf = objects.String(json_key='vrf')
    pid = objects.String(json_key='pid')
    start_time = objects.Float(json_key='startTime')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.name, cls.vrf)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.is_enabled, cls.is_active, cls.status, cls.is_monitored,
                cls.start_time, cls.pid,
                cls.timestamp, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Services, self).save_entry(connection, publish=True)


class OS(_RedisModel):
    name = objects.String(json_key='name')
    version = objects.String(json_key='version')
    version_id = objects.String(json_key='version_id')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.name, cls.version,
                cls.version_id)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(OS, self).save_entry(connection)


class CPU(_RedisModel):
    arch = objects.String(json_key='arch')
    nos = objects.String(json_key='nos')
    model = objects.String(json_key='model')
    max_freq = objects.String(json_key='max_freq')
    mem_total = objects.String(json_key='mem_total')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.arch, cls.nos, cls.model,
                cls.max_freq, cls.mem_total)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp,)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(CPU, self).save_entry(connection)


class ASIC(_RedisModel):
    vendor = objects.String(json_key='vendor')
    model = objects.String(json_key='model')
    model_id = objects.String(json_key='model_id')
    core_bw = objects.String(json_key='core_bw')
    ports = objects.String(json_key='ports')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.vendor, cls.model,
                cls.model_id, cls.core_bw, cls.ports)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp,)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(ASIC, self).save_entry(connection)


class Disk(_RedisModel):
    name = objects.String(json_key='name')
    size = objects.String(json_key='size')
    d_type = objects.String(json_key='d_type')
    vendor = objects.String(json_key='vendor')
    transport = objects.String(json_key='transport')
    rev = objects.String(json_key='rev')
    model = objects.String(json_key='model')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.name, cls.size, cls.d_type,
                cls.vendor, cls.transport, cls.rev, cls.model)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp,)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Disk, self).save_entry(connection)


class Board(_RedisModel):
    vendor = objects.String(json_key='vendor')
    model = objects.String(json_key='model')
    base_mac = objects.String(json_key='base_mac')
    part_number = objects.String(json_key='part_number')
    mfg_date = objects.String(json_key='mfg_date')
    serial_number = objects.String(json_key='serial_number')
    label_revision = objects.String(json_key='label_revision')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.vendor, cls.model,
                cls.base_mac, cls.part_number, cls.mfg_date, cls.serial_number,
                cls.label_revision)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp,)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Board, self).save_entry(connection)


class Memory(_RedisModel):
    vendor = objects.String(json_key='vendor')
    size = objects.String(json_key='size')
    speed = objects.String(json_key='speed')
    name = objects.String(json_key='name')
    m_type = objects.String(json_key='m_type')
    serial_number = objects.String(json_key='serial_number')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.name, cls.size, cls.speed,
                cls.m_type, cls.vendor, cls.serial_number)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp,)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Memory, self).save_entry(connection)

class Temp(_RedisModel):
    s_name = objects.String(json_key='name')
    s_desc = objects.String(json_key='desc')
    s_input = objects.Float(json_key='input')
    s_crit = objects.Integer(json_key='critical')
    s_lcrit = objects.Integer(json_key='lcritical')
    s_max = objects.Integer(json_key='high')
    s_min = objects.Integer(json_key='low')
    s_state = objects.String(json_key='state')
    s_prev_state = objects.String(json_key='prev_state')
    s_msg = objects.String(json_key='message')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.s_name,
                cls.s_desc)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.s_state, cls.s_prev_state, cls.s_input,
                cls.s_msg, cls.s_crit, cls.s_max, cls.s_min, cls.s_lcrit)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        if self.s_prev_state != self.s_state:
            super(Temp, self).save_entry(connection, publish=True)
        else:
            super(Temp, self).save_entry(connection, publish=False)


class Fan(_RedisModel):
    s_name = objects.String(json_key='name')
    s_desc = objects.String(json_key='desc')
    s_input = objects.Integer(json_key='input')
    s_max = objects.Integer(json_key='high')
    s_min = objects.Integer(json_key='low')
    s_state = objects.String(json_key='state')
    s_prev_state = objects.String(json_key='prev_state')
    s_msg = objects.String(json_key='message')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.s_name,
                cls.s_desc)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.s_state, cls.s_prev_state, cls.s_input,
                cls.s_msg, cls.s_max, cls.s_min)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        if self.s_prev_state != self.s_state:
            super(Fan, self).save_entry(connection, publish=True)
        else:
            super(Fan, self).save_entry(connection, publish=False)


class PSU(_RedisModel):
    s_name = objects.String(json_key='name')
    s_state = objects.String(json_key='state')
    s_prev_state = objects.String(json_key='prev_state')
    s_msg = objects.String(json_key='message')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.s_name

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return cls.timestamp, cls.s_state, cls.s_prev_state, cls.s_msg

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        if self.s_prev_state != self.s_state:
            super(PSU, self).save_entry(connection, publish=True)
        else:
            super(PSU, self).save_entry(connection, publish=False)

class EsxHost(_RedisModel):
    vcenter = objects.String(json_key='vcenter')
    datacenter = objects.String(json_key='datacenter')
    cluster = objects.String(json_key='cluster')
    esxhost = objects.String(json_key='esxhost')
    pnic = objects.String(json_key='pnic')
    dvs = objects.String(json_key='dvs')
    pg = objects.String(json_key='pg')
    encaptype = objects.Integer(json_key='encaptype')
    encap = objects.Integer(json_key='encap')
    vmname = objects.String(json_key='vmname')
    ip = objects.String(json_key='ip')
    mac = objects.String(json_key='mac')
    mtu = objects.Integer(json_key='mtu')
    pnic_admin_state = objects.String(json_key='pnic_admin_state')
    pnic_oper_state = objects.String(json_key='pnic_oper_state')
    peer_hostname = objects.String(json_key='peer_hostname')
    peer_ifname = objects.String(json_key='peer_ifname')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname,
                cls.vcenter, cls.datacenter, cls.cluster, cls.esxhost,
                cls.pnic, cls.dvs, cls.pg, cls.encaptype, cls.encap)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.ip, cls.mac, cls.mtu, cls.pnic_admin_state,
               cls.pnic_oper_state, cls.peer_hostname, cls.peer_ifname,
               cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(EsxHost, self).save_entry(connection)


class EsxVmAttr(_RedisModel):
    vcenter = objects.String(json_key='vcenter')
    datacenter = objects.String(json_key='datacenter')
    vmname = objects.String(json_key='vmname')
    ostype = objects.String(json_key='ostype')
    vnic = objects.String(json_key='vnic')
    pg = objects.String(json_key='portgroup')
    ip = objects.String(json_key='ip')
    mac = objects.String(json_key='mac')
    esxhost = objects.String(json_key='esxhost')
    mem = objects.String(json_key='memory')
    disk = objects.String(json_key='disk')
    cpus = objects.String(json_key='numcpus')
    vmpath = objects.String(json_key='vmpath')
    state = objects.String(json_key='powerstate')
    annotation = objects.String(json_key='annotation')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.vcenter, cls.datacenter,
                cls.vmname, cls.ostype, cls.vnic, cls.pg, cls.ip, cls.mac)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.esxhost, cls.mem, cls.disk, cls.cpus, cls.vmpath,
                cls.state, cls.annotation, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(EsxVmAttr, self).save_entry(connection)

class EsxVmBind(_RedisModel):
    vcenter = objects.String(json_key='vcenter')
    datacenter = objects.String(json_key='datacenter')
    cluster = objects.String(json_key='cluster')
    esxhost = objects.String(json_key='esxhost')
    vmname = objects.String(json_key='vmname')
    ostype = objects.String(json_key='ostype')
    state = objects.String(json_key='state')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.vcenter, cls.datacenter,
                cls.cluster, cls.esxhost, cls.vmname, cls.ostype, cls.state)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.state, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(EsxVmBind, self).save_entry(connection)

class DockerContainer(_RedisModel):
    container_id = objects.String(json_key='container_id')
    gw = objects.String(json_key='gw')
    image = objects.String(json_key='image')
    ip = objects.String(json_key='ip')
    mac = objects.String(json_key='mac')
    name = objects.String(json_key='name')
    network_name = objects.String(json_key='network_name')
    network_id = objects.String(json_key='network_id')
    prefix_len = objects.Integer(json_key='prefix_len')
    port_list = objects.String(json_key='port_list')
    service_id = objects.String(json_key='service_id')
    service_name = objects.String(json_key='service_name')
    status = objects.String(json_key='status')
    start_time = objects.String(json_key='start_time')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname,
                cls.container_id, cls.name, cls.image,
                cls.network_name, cls.ip, cls.mac,
                cls.service_name)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.status,
                cls.network_id, cls.gw, cls.prefix_len,
                cls.port_list, cls.service_id,
                cls.start_time, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(DockerContainer, self).save_entry(connection)

class DockerPortMap(_RedisModel):
    container_port = objects.String(json_key='container_port')
    container_id = objects.String(json_key='container_id')
    container_ip = objects.String(json_key='container_ip')
    host_port = objects.String(json_key='host_port')
    host_ip = objects.String(json_key='host_ip')
    image = objects.String(json_key='image')
    mac = objects.String(json_key='mac')
    name = objects.String(json_key='name')
    network_name = objects.String(json_key='network_name')
    network_id = objects.String(json_key='network_id')
    node_id = objects.String(json_key='node_id')
    proto = objects.String(json_key='proto')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname,
                cls.name, cls.container_ip, cls.proto, cls.container_port,
                cls.host_ip, cls.host_port, cls.network_name)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.container_id, cls.network_id,
                cls.image, cls.mac, cls.node_id, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(DockerPortMap, self).save_entry(connection)

class DockerSwarmCluster(_RedisModel):
    cluster_version = objects.String(json_key='cluster_version')
    cluster_name = objects.String(json_key='cluster_name')
    cluster_id = objects.String(json_key='cluster_id')
    docker_version = objects.String(json_key='docker_version')
    nodes = objects.String(json_key='nodes')
    num_nodes = objects.Integer(json_key='num_nodes')
    num_managers = objects.Integer(json_key='num_managers')
    managers = objects.String(json_key='managers')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.docker_version,
                cls.cluster_id, cls.cluster_name)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.cluster_version, cls.num_nodes, cls.num_managers, cls.managers,
                cls.timestamp, cls.nodes, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(DockerSwarmCluster, self).save_entry(connection)

class DockerSwarmNode(_RedisModel):
    cluster_id = objects.String(json_key='cluster_id')
    cluster_name = objects.String(json_key='cluster_name')
    docker_version = objects.String(json_key='docker_version')
    node_id = objects.String(json_key='node_id')
    node_name = objects.String(json_key='node_name')
    node_state = objects.String(json_key='node_state')
    plugins = objects.String(json_key='plugins')
    role = objects.String(json_key='role')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname,  cls.cluster_id, cls.cluster_name,
                cls.role, cls.node_state, cls.node_id, cls.node_name, cls.docker_version)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.plugins, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(DockerSwarmNode, self).save_entry(connection)

class DockerNetwork(_RedisModel):
    container_list = objects.String(json_key='container_list')
    driver = objects.String(json_key='driver')
    host_binding = objects.String(json_key='host_binding')
    ipv6_enabled = objects.Boolean(json_key='ipv6_enabled')
    ip_masquerade = objects.Boolean(json_key='ip_masquerade')
    ipam_driver = objects.String(json_key='ipam_driver')
    encrypted = objects.Boolean(json_key='encrypted')
    default_bridge = objects.Boolean(json_key='default_bridge')
    network_name = objects.String(json_key='network_name')
    network_id = objects.String(json_key='network_id')
    subnet = objects.String(json_key='subnet')
    gateway = objects.String(json_key='gateway')
    parent_interface = objects.String(json_key='parent_interface')
    vxlan_id = objects.Integer(json_key='vxlan_id')
    mtu = objects.Integer(json_key='mtu')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.network_name, cls.driver,
                cls.network_id, cls.subnet)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.gateway, cls.parent_interface, cls.vxlan_id,
               cls.mtu, cls.host_binding, cls.ipv6_enabled, cls.ip_masquerade,
               cls.encrypted, cls.default_bridge, cls.ipam_driver, cls.container_list,
               cls.timestamp, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(DockerNetwork, self).save_entry(connection)


class DockerService(_RedisModel):
    image = objects.String(json_key='image')
    mode = objects.String(json_key='mode')
    parallelism = objects.Integer(json_key='parallelism')
    port_list = objects.String(json_key='port_list')
    network_list = objects.String(json_key='network_list')
    replicas = objects.Integer(json_key='replicas')
    service_name = objects.String(json_key='service_name')
    service_id = objects.String(json_key='service_id')
    version = objects.Integer(json_key='version')
    vip = objects.String(json_key='vip')


    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.service_id,
                cls.service_name, cls.mode)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.image, cls.replicas, cls.parallelism,
               cls.port_list, cls.network_list, cls.vip, cls.version,
               cls.timestamp, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(DockerService, self).save_entry(connection)


class DockerHost(_RedisModel):
    containers = objects.Integer(json_key='containers')
    containers_running = objects.Integer(json_key='containers_running')
    docker_version = objects.String(json_key='version')
    images = objects.Integer(json_key='image_count')
    ip_forwarding = objects.Boolean(json_key='ip_forwarding')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.docker_version)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.images, cls.containers, cls.containers_running,
                cls.ip_forwarding, cls.timestamp, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(DockerHost, self).save_entry(connection)


class License(_RedisModel):
    name = objects.String(json_key='name')
    state = objects.String(json_key='state')
    license = objects.String(json_key='license')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.name

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return cls.state, cls.license, cls.timestamp

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(License, self).save_entry(connection, publish=True)
